#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Define a function to configure the logging module.
"""

import os
import logging.config

import yaml


def setup_logging(
        default_path=os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            'logging.yml'),
        default_level=logging.INFO
):
    """
    Setup logging configuration.
    """
    path = default_path
    try:
        if os.path.exists(path):
            with open(path, 'rt') as file:
                config = yaml.safe_load(file.read())
            logging.config.dictConfig(config)
            return "logging is using a yaml config file"
    except TypeError:
        logging.basicConfig(level=default_level)
        return "logging is using basic config"



#!/usr/bin/env python3
# -*- coding=utf-8 -*-

"""S3 analyst for Coveo Devops challenge. This tool will crunch metadata from \
S3 storage, store them into a local csv file and query over it. You can share \
your csv file with ease, automate its build, or re-use it again and again.

Usage:
  cli.py [<buckets>...] [--exclude-mine] [--fmtsize FMTSIZE] \
[--regex=<filter>] [(--block_data_crunching | \
[--force_update --default_timeout=<x>])] [[--path=<abs_path>] \
[--filename=<filename>] | --filepath=<full_filename>]
  cli.py [<buckets>...]
  cli.py [<buckets>...] [--exclude-mine]
  cli.py [<buckets>...] [--exclude-mine] [--regex=<filter>]
  cli.py [<buckets>...] [--exclude-mine] [--fmtsize FMTSIZE]
  cli.py [<buckets>...] [--exclude-mine] [(--block_data_crunching | \
[--force_update --default_timeout=<x>])]
  cli.py [<buckets>...] [--regex=<filter>] [[--path=<abs_path>] \
[--filename=<filename>] | --filepath=<full_filename>]

Examples:
  >>> # Crunch data from aws public dataset [without yours for genericity]
  >>> cli.py dc-lidar-2015 --exclude-mine --force_update
  >>> # Crunch data quicker from aws public dataset (using cache)
  >>> cli.py dc-lidar-2015 -e
  >>> # Be the force with you and crunch data from aws public dataset \
(ignore cache)
  >>> cli.py dc-lidar-2015 -e --force_update
  >>> # Filter metadata with regex and avoid to crunch metadata again \
(force using cache)
  >>> cli.py -e --block_data_crunching --regex='2011'
  >>> # Shortest filter metadata with regex (force using cache with -b option)
  >>> cli.py -eb --regex='2011'
  >>> # An example of a little more complex regex
  >>> cli.py -eb --regex='^dc-lidar-2015/Classified_LAS/11[a-z]*'
  >>> # You can change the size unit
  >>> cli.py -eb --fmtsize TB
  >>> # Wow, this tool begins to be impressive, isn't it?
  >>> # Let's go deeper in foolish!
  >>> # You can also get informations of multiple buckets. \
[Here is another public dataset hosted on aws] \
[NB: about 1 minute is required depending how your Internet is.]
  >>> cli.py -e dc-lidar-2015 fmi-opendata-rcrhirlam-surface-grib \
--force_update
  >>> # That's a lot of data !
  >>> # Now, it's time to work together.
  >>> # Let's crunch the data again in another csv file.
  >>> cli.py dc-lidar-2015 --exclude-mine --path="/home/${USER}" \
--filename='another_cache.csv'
  >>> # Or more consise way
  >>> cli.py dc-lidar-2015 --exclude-mine \
--filepath="/home/${USER}/another_cache.csv"
  >>> # Let's play or check the documentation on the Internet for more options
  >>> cli.py --help

Options:
  <buckets>                     The name of the buckets to analyze.

  -h, --help                    Show this screen.
  -v, --version                 Show version.
  --fmtsize unit                Format the display of the size in the results.
                                    Leave blank for auto-choosen unit.
                                    Unit should be one of :
                                    '', 'B', 'KB', 'KiB', 'MB', 'MiB', 'GB',
                                    'GiB', 'TB', 'TiB', 'PB', 'PiB', 'EB',
                                    'EiB', 'ZB', 'ZiB'
                                    [default: ].
  -e, --exclude-mine            Do not consider your own buckets.
                                    [default: False].
  --regex=<filter>              Regex to filter the bucket output.
  -b, --block_data_crunching    Avoid to collect data from the Internet.
  -u, --force_update            Force to collect data from the Internet.
  --default_timeout=<x>         Set an amount of seconds to consider the
                                    collected data file as expired.
                                    [default: 60]
  --path=<abs_path>             Full path of directory to store data.
  --filename=<filename>         Name of the file to store data.
  --filepath=<full_filename>    Name of the file to store data. Path included.
"""

import logging
from docopt import docopt

try:
    from s3_analyst.__init__ import __version__
    from s3_analyst import s3_analyst, s3_analyst_logging
except ImportError:  # pragma: no cover
    import s3_analyst
    import s3_analyst_logging
    from __init__ import __version__

from schema import Schema, SchemaError, Or

# configure the logger
s3_analyst_logging.setup_logging()
LOGGER = logging.getLogger(__name__)


def validate(opts):
    """
    Check the opts inputs to respect a specific format.
    """
    schema = Schema({
        '<buckets>':  Or(None, str, [str]),
        '--fmtsize':   Or(None, '', 'B', 'KB', 'KiB', 'MB', 'MiB', 'GB',
                          'GiB', 'TB', 'TiB', 'PB', 'PiB', 'EB', 'EiB',
                          'ZB', 'ZiB',
                          error='fmtsize option should respect the format : '
                                'B, KB, KiB, MB, MiB, GB, GiB, TB, TiB, PB, '
                                'PiB, EB, EiB, ZB, ZiB.'),
        '--exclude-mine': bool,
        '--block_data_crunching': bool,
        '--force_update': bool,
        '--regex': Or(None, str),
        '--default_timeout': int,
        '--path': Or(None, str),
        '--filename': Or(None, str),
        '--filepath': Or(None, str),
    })
    try:
        args = schema.validate(opts)
        return args
    except SchemaError:
        raise


def parser(args):
    """
    Validate args and call some conversion functions.
    """
    # type conversion for validation
    args['--default_timeout'] = int(args['--default_timeout'])

    try:
        validate(args)
    except Exception as error:
        raise SchemaError('%r raised %r' % (validate.__name__, error))

    # type conversion for module
    bucket_list = s3_analyst.str_to_bucket_list(
        args['<buckets>'],
        own_buckets=not args['--exclude-mine'])

    path = args['--path']
    filename = args['--filename']
    filepath = args['--filepath']
    if filepath is None:
        filepath = s3_analyst.default_csv_filepath(path, filename)

    args['bucket_list'] = bucket_list
    args['filepath'] = filepath

    return args


def main(argv=None):
    """
    Decode inputs, call parser and call s3_analyst.main.
    """
    arguments = docopt(__doc__, argv, version=__version__)
    args = parser(arguments)

    kwargs = {}
    kwargs['bucket_list'] = args['bucket_list']
    kwargs['block_data_crunching'] = args['--block_data_crunching']
    kwargs['force_update'] = args['--force_update']
    kwargs['regex_filtering'] = args['--regex']
    kwargs['fmt_size'] = args['--fmtsize']
    kwargs['default_timeout'] = args['--default_timeout']
    kwargs['filepath'] = args['filepath']

    LOGGER.info("cli.main kwargs: %s", kwargs)

    out = s3_analyst.main(**kwargs)

    return out


if __name__ == '__main__':  # pragma: no cover
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `s3_analyst` package."""

import pytest

from hypothesis import given
from hypothesis.strategies import just, integers, floats, text
from freezegun import freeze_time

import boto3
from moto import mock_s3

from s3_analyst import s3_analyst


@pytest.yield_fixture(scope="function")
def fake_s3():
    """
    A pytest fixture to simulate aws s3 storage behaviours.
    """
    mock_s3().start()

    class FakeS3(object):
        """
        A fake s3 storage which contains calls to client and resource.
        """
        def __init__(self, file=None):
            self.client = boto3.client("s3")
            self.resource = boto3.resource("s3")
            self.file = str(file)

        def false_bucket(self, name):
            """
            Create a false bucket.
                :name: Given name to the bucket.
            """
            self.resource.create_bucket(Bucket=name)

        def fill_bucket(self, bucket_name, bucket_key):
            """
            Add the content of self.file to a new key into the bucket.
            """
            bucket = self.resource.Bucket(bucket_name)
            with open(self.file, 'rb') as body:
                bucket.put_object(Key=bucket_key, Body=body)

    yield FakeS3

    mock_s3().stop()


@pytest.fixture()
def no_user_dependant(monkeypatch, tmpdir):
    """
    Replace os.path.expanduser by a directory created for the test.
    """
    import os.path

    def mockreturn(_):
        """
        Mock expanduser.
        """
        return str(tmpdir)
    monkeypatch.setattr(os.path, 'expanduser', mockreturn)
    return str(tmpdir)


@pytest.fixture(scope="function")
def a_file(tmpdir):
    """
    Generate a file in a structured format for the test.
    """
    test = tmpdir.join('file.csv')
    s3_analyst.init_csv_file(str(test))
    return test


@pytest.fixture()
def no_time_dependant(monkeypatch):
    """
    Replace os.path.getctime by a fixed timestamp for the test.
    """
    import os.path
    from datetime import datetime

    a_datetime = datetime(2010, 11, 12, 13, 14, 15)

    def mockreturn(_):
        """
        Mock getctime.
        """
        return a_datetime.timestamp()

    monkeypatch.setattr(os.path, 'getctime', mockreturn)
    return a_datetime


def test_collect_informations():
    """
    Test s3_analyst.collect_informations with a csv sample file.
    """
    import os
    current_file = os.path.abspath(os.path.dirname(__file__))
    csv_filepath = os.path.join(current_file, 'data/aws_s3_export_2.csv')

    buckets_info = s3_analyst.collect_informations(csv_filepath)

    assert buckets_info == [
        {
            'creation date': '2017-09-16 08:59:36',
            'object': 1,
            'size': 81897,
            'location': 'eu-central-1',
            'last modified date': '2017-09-16 09:12:11',
            'bucket name': 'testcoveoprotony'
        }, {
            'creation date': '2017-09-16 08:58:36',
            'object': 3,
            'size': 90671,
            'location': 'eu-central-1',
            'last modified date': '2017-09-16 09:02:11',
            'bucket name': 'testcoveotofull'}]


@pytest.mark.parametrize("regex_param", [
    "^testcoveop[a-z]*",
    "^[a-z]*protony[a-z]*",
    "proto",
    "[a-z]*y"
    ])
def test_collect_info_regex_filter(regex_param):
    """
    Test collect_informations with a csv sample file and filter option. \
    Filter are designed to return information about \
    the testcoveoprotony bucket name.
    """
    import os
    current_file = os.path.abspath(os.path.dirname(__file__))
    csv_filepath = os.path.join(current_file, 'data/aws_s3_export_2.csv')

    regex_filter = regex_param
    buckets_info = s3_analyst.collect_informations(csv_filepath, regex_filter)
    assert buckets_info == [
        {
            'creation date': '2017-09-16 08:59:36',
            'object': 1,
            'size': 81897,
            'location': 'eu-central-1',
            'last modified date': '2017-09-16 09:12:11',
            'bucket name': 'testcoveoprotony'
        }]


def test_init_csv_file(tmpdir):
    """
    Test if the new csv output file has headers as expected.
    """
    csv_file = tmpdir.join('output.csv')
    s3_analyst.init_csv_file(str(csv_file))
    with open(str(csv_file), 'rU') as file:
        assert file.read() == ('"bucket name","location","creation date",'
                               '"object","last modified date","owner","size",'
                               '"storage class"\n')


@given(floats())
def test_sizeof_fmt_floats(bytes_size):
    """
    Test the sizeof_fmt function with floats.
    """
    output = s3_analyst.sizeof_fmt(bytes_size, forced_unit='B')
    assert output == '{:3.1f}'.format(bytes_size)+'B'


@given(integers(max_value=9007199254740992))
def test_sizeof_fmt_integers(bytes_size):
    """
    Test the sizeof_fmt function with integers.

    If bytes_size is an integer, it should be a "valid" integers, \
    ie. it should respect the IEEE 754 standard. That means some problems \
    could occur because, any number requiring more than 52 significant \
    bits is rounded to 52 significant bits. And 2^53 = 9007199254740992.
    `See more details. <https://stackoverflow.com/q/3793838>`_
    """
    output = s3_analyst.sizeof_fmt(bytes_size, forced_unit='B')
    assert output == str(bytes_size)+'.0B'


@given(integers(min_value=9007199254740993))
def test_sizeof_fmt_invalid_int(bytes_size):
    """
    Test the sizeof_fmt function with invalid integers.

    If bytes_size is an integer, it should be a "valid" integers, \
    ie. it should respect the IEEE 754 standard. That means some problems \
    could occur because, any number requiring more than 52 significant \
    bits is rounded to 52 significant bits. And 2^53 = 9007199254740992.
    `See more details. <https://stackoverflow.com/q/3793838>`_
    """
    with pytest.raises(ArithmeticError):
        s3_analyst.sizeof_fmt(bytes_size, forced_unit='B')


@pytest.mark.parametrize("bytes_size,unit,expected", [
    (0, 'B', "0.0B"),
    (0, 'MB', "0.0MB"),
    (1000, 'KB', "1.0KB"),
    (1024, 'KiB', "1.0KiB"),
    (9000, 'KB', "9.0KB"),
    (9234, 'KB', "9.2KB"),
    (9734, 'KB', "9.7KB"),
    (10000, 'KB', "10.0KB"),
    (100000, 'KB', "100.0KB"),
    (1000000, 'KB', "1000.0KB"),
    (1000000, 'KiB', "976.6KiB"),
    (1000000, 'MB', "1.0MB"),
    (10000000, 'KB', "10000.0KB"),
    (10000000, 'MB', "10.0MB"),
    (1000000000, 'GB', "1.0GB"),
    (1000000000000, 'TB', "1.0TB"),
    (1000000000000000, 'PB', "1.0PB"),
    (1000000000000000000, 'EB', "1.0EB"),
    (100000000000000000000, 'ZB', "0.1ZB"),
    (1000000000000000000000, 'ZB', "1.0ZB"),
    (1000000000000000000000, 'KB', "1000000000000000000.0KB"),
])
def test_sizeof_fmt_param_unit(bytes_size, unit, expected):
    """
    Test sizeof_fmt given a forced_unit.
    """
    output = s3_analyst.sizeof_fmt(bytes_size, forced_unit=unit)
    assert output == expected


@pytest.mark.parametrize("bytes_size,expected_output", [
    (0, "0.0B"),
    (1000, "1.0KB"),
    (1024, "1.0KB"),
    (9000, "9.0KB"),
    (9234, "9.2KB"),
    (9734, "9.7KB"),
    (10000, "10.0KB"),
    (100000, "100.0KB"),
    (409600, "409.6KB"),
    (1000000, "1.0MB"),
    (1000000, "1.0MB"),
    (10000000, "10.0MB"),
    (409600425, "409.6MB"),
    (4096004250, "4.1GB"),
    (1000000000, "1.0GB"),
    (1000000000000, "1.0TB"),
    (1000000000000000, "1.0PB"),
    (1000000000000000000, "1.0EB"),
    (100000000000000000000, "100.0EB"),
    (1000000000000000000000, "1.0ZB"),
])
def test_sizeof_fmt_parametrized(bytes_size, expected_output):
    """
    Test sizeof_fmt given an empty forced_unit (auto-choosen unit mode).
    """
    output = s3_analyst.sizeof_fmt(bytes_size, forced_unit='')
    assert output == expected_output


@given(integers(
    max_value=9007199254740992),
       text().filter(lambda txt:
                     txt not in [''] +
                     ['B', 'KB', 'MB', 'GB',
                      'TB', 'PB', 'EB', 'ZB'] +
                     ['B', 'KiB', 'MiB', 'GiB',
                      'TiB', 'PiB', 'EiB', 'ZiB']))
def test_sizeof_fmt_invalid_unit(bytes_size, forced_unit):
    """
    Test sizeof_fmt given an invalid unit. Should return the output without \
    change but in B.
    """
    output = s3_analyst.sizeof_fmt(bytes_size, forced_unit=forced_unit)
    assert output == str(bytes_size)+'.0B'


def test_format_output():
    """
    Test format_output with minimal input.
    """
    arg = {'size': 0}
    out = s3_analyst.format_output(arg)
    assert out == {'size': '0.0B', 'filter': None}


def test_format_output_filter():
    """
    Test format_output with minimal input and filter option.
    """
    arg = {'size': 0}
    regex_filter = '^[a-z]*'
    out = s3_analyst.format_output(arg, filter=regex_filter)
    assert out == {'size': '0.0B', 'filter': regex_filter}


def test_is_file_expired_notexist(tmpdir):
    """
    Test if a file is expired when it does not exist.
    """
    file = tmpdir.join('notexist.csv')
    out = s3_analyst.is_file_expired(str(file), threshold=60)
    assert out


def test_is_expired_just_created(tmpdir):
    """
    Test if a file is expired when threshold = 0.
    """
    file = tmpdir.join('output.csv')
    s3_analyst.init_csv_file(str(file))
    out = s3_analyst.is_file_expired(str(file), threshold=0)
    assert out


@pytest.mark.parametrize("seconds,threshold,expected", [
    (10, 60, False),  # "50 seconds until expired"
    (61, 60, True),  # "expired from 1 second"
    (59, 60, False),  # "1 second until expired"
    (1, 0, True),  # "expired from 1 second"
    (10, 0, True),  # "expired from 10 seconds"
    (0, 0, False),  # "get the limit"
    (60, 60, False),  # "get the limit"
    (0, 1, False),  # "1 second until expired"
])
def test_is_file_expired(tmpdir, no_time_dependant, seconds,
                         threshold, expected):
    """
    Test if a file is expired with travelling in time.
    """
    from datetime import timedelta

    # Reduce non-determinism properties
    a_datetime = no_time_dependant
    a_timedelta = timedelta(seconds=seconds)
    fake_datetime = a_datetime + a_timedelta

    # Create file at date a_datetime
    with freeze_time(a_datetime):
        file = tmpdir.join('output.csv')
        s3_analyst.init_csv_file(str(file))

    # Check if it is expired at a_datetime + a_timedelta
    with freeze_time(fake_datetime):
        out = s3_analyst.is_file_expired(str(file), threshold=threshold)
        assert out == expected


def test_str_to_bucket_list_empty():
    """
    Test the conversion from a empty str list to bucket list.
    """
    out = s3_analyst.str_to_bucket_list(string_list=[], own_buckets=False)
    assert out == []


def test_str_to_bucket_list_none():
    """
    Test the conversion from a None str list to bucket list.
    """
    out = s3_analyst.str_to_bucket_list(string_list=None, own_buckets=False)
    assert out == []


@given(just(['firstbucket', 'secondbucket']))
def test_str_to_bucket_list(fake_s3, bucket_list):
    """
    Test the conversion from a str list to bucket list.
    """
    s3_resource = fake_s3().resource
    out = s3_analyst.str_to_bucket_list(
        string_list=bucket_list, own_buckets=False)
    assert out == list(map(lambda bucket_name:
                           s3_resource.Bucket(name=bucket_name),
                           bucket_list))


def test_crunch_data(fake_s3, a_file):
    """
    Test if crunch_data write all the information from the fake s3 storage.
    """
    from datetime import datetime
    a_datetime = datetime(2010, 11, 12, 13, 14, 15)

    # Reduce non-determinism properties
    with freeze_time(a_datetime):
        # Configure the fake s3
        empty_s3 = fake_s3(a_file)
        empty_s3.false_bucket('firstbucket')
        empty_s3.fill_bucket('firstbucket', 'falseBucket.test')
        a_bucket = empty_s3.resource.Bucket('firstbucket')

        # Crunch data and compare with the expected output
        s3_analyst.crunch_data(str(a_file), bucket_list=[a_bucket])

        mtime_fmt = a_datetime.strftime('%Y-%m-%d %H:%M:%S')
        expected_output = (
            '"bucket name","location","creation date","object",'
            '"last modified date","owner","size","storage class"\n'
            '"firstbucket","us-east-1","2006-02-03 16:45:09",'
            '"firstbucket/falseBucket.test","%s",'
            '"75aa57f09aa0c8caeab4f8c24e99d10f8e7faeebf76c078efc7c6caea54ba06a'
            '","103","STANDARD"\n') % mtime_fmt
        with open(str(a_file), 'rU') as file:
            assert file.read() == expected_output


def test_main(fake_s3, a_file):
    """
    Test if the main function write all the information from the fake \
    s3 storage and return the information.
    """
    from datetime import datetime
    a_datetime = datetime(2010, 11, 12, 13, 14, 15)

    # Reduce non-determinism properties
    with freeze_time(a_datetime):
        # Configure the fake s3
        empty_s3 = fake_s3(a_file)
        empty_s3.false_bucket('firstbucket')
        empty_s3.fill_bucket('firstbucket', 'falseBucket.test')
        a_bucket = empty_s3.resource.Bucket('firstbucket')

        # Crunch data and compare with the expected output
        out = s3_analyst.main(**dict(
            bucket_list=[a_bucket],
            force_update=True,
            filepath=str(a_file)))

        mtime_fmt = a_datetime.strftime('%Y-%m-%d %H:%M:%S')
        expected_output = (
            '"bucket name","location","creation date","object",'
            '"last modified date","owner","size","storage class"\n'
            '"firstbucket","us-east-1","2006-02-03 16:45:09",'
            '"firstbucket/falseBucket.test","%s",'
            '"75aa57f09aa0c8caeab4f8c24e99d10f8e7faeebf76c078efc7c6caea54ba06a'
            '","103","STANDARD"\n'
            ) % mtime_fmt

        with open(str(a_file), 'rU') as file:
            assert file.read() == expected_output

        expected_output = [{
            'location': 'us-east-1',
            'filter': '^[a-z]*',
            'last modified date': mtime_fmt,
            'size': '103.0B',
            'creation date': '2006-02-03 16:45:09',
            'bucket name': 'firstbucket',
            'object': 1}]
        assert out == expected_output



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests the s3_analyst_logging part of the s3_analyst package."""

from s3_analyst import s3_analyst_logging


def test_setup_logging_default():
    """
    Test s3_analyst_logging.setup_logging with the default argument.
    Should use a predefined yaml config file.
    """
    output = s3_analyst_logging.setup_logging()
    assert output == 'logging is using a yaml config file'


def test_setup_logging_basic_config():
    """
    Test s3_analyst_logging.setup_logging with a bad config path.
    Should use the basic config.
    """
    output = s3_analyst_logging.setup_logging(None)
    assert output == 'logging is using basic config'



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests the cli of the s3_analyst package."""

from string import ascii_letters

import pytest

from s3_analyst import cli

import boto3
from moto import mock_s3

from hypothesis import given
from hypothesis.strategies import just, lists, text
from freezegun import freeze_time


@pytest.yield_fixture(scope="function")
def fake_s3():
    """
    A pytest fixture to simulate aws s3 storage behaviours.
    """
    mock_s3().start()

    class FakeS3(object):
        """
        A fake s3 storage which contains calls to client and resource.
        """
        def __init__(self, file=None):
            self.client = boto3.client("s3")
            self.resource = boto3.resource("s3")
            self.file = str(file)

        def false_bucket(self, name):
            """
            Create a false bucket.
                :name: Given name to the bucket.
            """
            self.resource.create_bucket(Bucket=name)

        def fill_bucket(self, bucket_name, bucket_key):
            """
            Add the content of self.file to a new key into the bucket.
            """
            bucket = self.resource.Bucket(bucket_name)
            with open(self.file, 'rb') as body:
                bucket.put_object(Key=bucket_key, Body=body)

    yield FakeS3

    mock_s3().stop()


@pytest.fixture()
def no_user_dependant(monkeypatch, tmpdir):
    """
    Replace os.path.expanduser by a directory created for the test.
    """
    import os.path

    def mockreturn(_):
        """
        Mock expanduser.
        """
        return str(tmpdir)
    monkeypatch.setattr(os.path, 'expanduser', mockreturn)
    return str(tmpdir)


@pytest.fixture(scope="function")
def a_file(tmpdir):
    """
    Generate a file in a structured format for the test.
    """
    test = tmpdir.join('file.csv')
    from s3_analyst import s3_analyst
    s3_analyst.init_csv_file(str(test))
    return test


@given(lists(text(ascii_letters, min_size=1), min_size=1))
def test_cli_bucket_option(capsys, bucket_list):
    """
    Test if the cli recognizes the bucket option.
    """
    try:
        if isinstance(bucket_list[0], unicode):  # noqa
            new_bucket_list = []
            for item in bucket_list:
                new_bucket_list.append(item.encode('utf-8', 'ignore'))
            bucket_list = new_bucket_list
    except Exception:
        # hypothesis.strategies.text generates str on Python3, unicode on 2.7.
        # docopt fails to read repeating elements if unicode is used
        # https://github.com/docopt/docopt/issues/301
        # https://github.com/docopt/docopt/issues/219
        pass
    from docopt import docopt

    doc = cli.__doc__
    args = docopt(doc, bucket_list)
    out, _ = capsys.readouterr()
    print(out)
    assert args["<buckets>"] == bucket_list


@given(just(['firstbucket', 'secondbucket']))
def test_cli_parser_bucket_list(no_user_dependant, fake_s3, bucket_list):
    """
    Test if cli.parser returns the list of buckets.
    """
    s3_resource = fake_s3().resource

    from docopt import docopt
    doc = cli.__doc__
    args = docopt(doc, bucket_list)
    result = cli.parser(args)

    assert result == {
        '--block_data_crunching': False,
        '--default_timeout': 60,
        '--exclude-mine': False,
        '--filename': None,
        '--filepath': None,
        '--fmtsize': '',
        '--force_update': False,
        '--path': None,
        '--regex': None,
        '<buckets>': bucket_list,
        'bucket_list': list(map(lambda bucket_name:
                                s3_resource.Bucket(name=bucket_name),
                                bucket_list)),
        'filepath': no_user_dependant+'/aws_s3_export.csv'}


@pytest.mark.parametrize("fmt_size", ['B',
                                      'KB', 'MB', 'GB',
                                      'KiB', 'MiB', 'GiB',
                                      'TB', 'PB', 'EB', 'ZB',
                                      'TiB', 'PiB', 'EiB', 'ZiB'])
def test_cli_parser_fmtsize(no_user_dependant, fake_s3, fmt_size):
    """
    Test if the cli.parser prints the message as expected.
    """
    fake_s3 = fake_s3().resource

    from docopt import docopt
    doc = cli.__doc__
    args = docopt(doc, ["firstbucket", '--exclude-mine',
                        '--fmtsize', fmt_size])
    out = cli.parser(args)
    expected_output = {
        '--block_data_crunching': False,
        '--default_timeout': 60,
        '--exclude-mine': True,
        '--filename': None,
        '--filepath': None,
        '--fmtsize': fmt_size,
        '--force_update': False,
        '--path': None,
        '--regex': None,
        '<buckets>': ['firstbucket'],
        'bucket_list': [fake_s3.Bucket(name='firstbucket')],
        'filepath': no_user_dependant+'/aws_s3_export.csv'}
    assert out == expected_output


def test_cli_parser_invalid_option():
    """
    Test if the cli.parser raises a schema error if an option is invalid.
    """
    from docopt import docopt
    from schema import SchemaError
    doc = cli.__doc__
    args = docopt(doc, ["firstbucket", '--fmtsize', 'To'])
    with pytest.raises(SchemaError):
        cli.parser(args)


def test_cli_parser_validation():
    """
    Test if the cli.parser prints a user-friendly message on failure.
    """
    from docopt import docopt
    from schema import SchemaError
    expected_output = ("'validate' raised SchemaError('fmtsize option should "
                       "respect the format : B, KB, KiB, MB, MiB, GB, GiB, "
                       "TB, TiB, PB, PiB, EB, EiB, ZB, ZiB.',)")
    with pytest.raises(SchemaError):
        try:
            doc = cli.__doc__
            args = docopt(doc, ["firstbucket", '--fmtsize', 'Mo'])
            cli.parser(args)
        except SchemaError as err:
            assert err.args[0] == expected_output
            raise


def test_cli_main(no_user_dependant, fake_s3):
    """
    Test if the cli.main exits when no arg.
    Note: sys.argv should not have any extra option after name of the test.
    """
    with pytest.raises((EOFError, SystemExit)):
        # raise EOFError because bucket list is empty.
        # raise Docopt.DocoptExit if sys.argv contains invalid option.
        cli.main()


def test_cli_main_help(capsys):
    """
    Test cli.main with help in sys.argv.
    """
    import sys
    doc = cli.__doc__
    sys.argv.append("-h")
    with pytest.raises(SystemExit):
        cli.main()
    out, _ = capsys.readouterr()
    assert out == doc


def test_cli_main_output(fake_s3, a_file):
    """
    Test how cli.main reacts with a list of args.
    """
    from datetime import datetime
    a_datetime = datetime(2010, 11, 12, 13, 14, 15)

    # Reduce non-determinism properties
    with freeze_time(a_datetime):

        # Configure the fake s3
        empty_s3 = fake_s3(a_file)
        empty_s3.false_bucket('firstbucket')
        empty_s3.fill_bucket('firstbucket', 'falseBucket.test')

        arguments = ["firstbucket", '--fmtsize', 'KB',
                     '--filepath', str(a_file), '--force_update']
        out = cli.main(arguments)

        mtime_fmt = a_datetime.strftime('%Y-%m-%d %H:%M:%S')
        # modif_time_fmt = datetime.utcfromtimestamp(
        #     a_file.mtime()).strftime('%Y-%m-%d %H:%M:%S')

        expected_output = [{
            'object': 1, 'bucket name': 'firstbucket', 'filter': None,
            'location': 'us-east-1', 'creation date': '2006-02-03 16:45:09',
            'size': '0.1KB', 'last modified date': mtime_fmt}]
        assert out == expected_output



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
A module which contains functions to crunch aws s3 metadata
and query over them.
"""

import csv
import datetime
import logging

import os.path
from os import error

import boto3
from botocore.exceptions import ClientError

import pandas as pd

try:
    from s3_analyst import s3_analyst_logging
except ImportError:  # pragma: no cover
    import s3_analyst_logging

# configure the logger
s3_analyst_logging.setup_logging()
LOGGER = logging.getLogger(__name__)

# declare aws variables
S3 = boto3.resource("s3")
S3_CLIENT = boto3.client("s3")


def default_csv_path(directory=None):
    """
    Return the given directory or set up a default one.
        :directory=expanduser("~"): String - given absolute directory.
    """
    if directory is None:
        directory = os.path.expanduser("~")
    return directory


def default_csv_filename(filename='aws_s3_export.csv'):
    """
    Return the given filename or set up a default one.
        :param filename='aws_s3_export.csv': String - given filename
    """
    return filename


def default_csv_filepath(
        path=None,
        filename=None):
    """
    Return the filepath given the path and the filename.
        :param path=default_csv_path(): String - given absolute directory
        :param filename=default_csv_filename(): String - given filename

    """
    if path is None:
        path = default_csv_path()
    if filename is None:
        filename = default_csv_filename()
    filepath = os.path.join(path, filename)
    return filepath


def init_csv_file(filepath=default_csv_filepath()):
    """
    Write headers of a csv file.\
    Return the filepath.
        :filepath=default_csv_filepath(): the absolute csv filepath
    """
    headers = [
        "bucket name",
        "location",
        "creation date",
        "object",
        "last modified date",
        "owner",
        "size",
        "storage class"]  # TODO(tofull) Pass this list as args

    with open(filepath, 'w') as file:
        wrt = csv.writer(file, quoting=csv.QUOTE_ALL)
        wrt.writerow(headers)

    return filepath


def crunch_data(filepath, bucket_list=None):
    """
    Crunch data and fill the csv file given by the filepath with informations
    from a list of buckets.
        :param filepath: Absolute path for csv file.
        :param bucket_list=None: The list of buckets.
    """

    bucket_info = [0 for _ in range(8)]

    with open(filepath, 'a') as export:
        wrt = csv.writer(export, quoting=csv.QUOTE_ALL)

        for bucket in bucket_list:
            try:
                location = S3_CLIENT.get_bucket_location(
                    Bucket=bucket.name
                )['LocationConstraint']
            except ClientError as cli_err:  # pragma: no cover
                LOGGER.error("Received error: %s", cli_err, exc_info=True)
                location = 'Unabled to find'

            try:
                bucket_creation_date = bucket.creation_date.strftime(
                    "%Y-%m-%d %H:%M:%S")
            except AttributeError as att_err:  # pragma: no cover
                LOGGER.error("Received error: %s", att_err, exc_info=True)
                bucket_creation_date = 'Unable to find'

            prefix = bucket.name + '/'
            for key in bucket.objects.all():
                # bucket.objects.all() returns a list of s3.objectSummary items
                time_format = key.last_modified
                formated_time = time_format.strftime("%Y-%m-%d %H:%M:%S")

                bucket_info[0] = bucket.name
                bucket_info[1] = location
                bucket_info[2] = bucket_creation_date
                bucket_info[3] = prefix + key.key
                bucket_info[4] = formated_time
                bucket_info[5] = key.owner['ID']
                bucket_info[6] = key.size
                bucket_info[7] = key.storage_class
                wrt.writerow(bucket_info)


def is_file_expired(filepath, threshold=60):
    """
    Return True if the file was created before a certain amount of time.
    Else otherwise.
        :param filepath: Absolute path to the file.
        :param threshold=60: Number of seconds.
    """
    try:
        date = os.path.getctime(filepath)
    except error:
        LOGGER.debug("Unable to get the creation date of %s.", filepath)
        return True
    now = datetime.datetime.now().timestamp()
    # nom = time.time()  # for python2 but bug with freeze_time in tests.
    # datetime.timestamp does not exist on python2...
    # >> This package does not support python2 anymore.
    ellapsed = now - date
    return ellapsed > threshold


def collect_informations(filepath, regex_filtering="^[a-z]*"):
    """
    Extract informations from csv file where bucket data are stored.
        :param filepath: Absolute filepath of the csv file
        :param regex_filtering: Regex to filter the result
    """

    try:
        data = pd.read_csv(filepath)
    except pd.errors.ParserError as pderror:
        raise EOFError('%s\n\nYour csv file may be empty. '
                       'Please try with the --force_update option or check '
                       'your bucket list.' % pderror)
    try:
        out = data.loc[data.loc[:, 'object'].str.contains(regex_filtering)]
    except TypeError:
        out = data

    out = out.groupby(['bucket name'], as_index=False).agg({
        'creation date': 'unique',
        'location': 'unique',
        'object': 'count',
        'size': 'sum',
        'last modified date': 'max'})

    buckets_info = []
    for bucket_index in range(len(out['bucket name'])):
        info = {
            'bucket name': out['bucket name'][bucket_index],
            'creation date': out['creation date'][bucket_index][0],
            'location': out['location'][bucket_index][0],
            'object': out['object'][bucket_index],
            'size': out['size'][bucket_index],
            'last modified date': out['last modified date'][bucket_index]
        }
        buckets_info.append(info.copy())

    return buckets_info


def sizeof_fmt(bytes_size, forced_unit='B'):
    """
    Convert the size into a more human readable format.
    The final unit could be forced by the second param.
        :param bytes_size: Size to convert (in bytes)
        :param forced_unit='B': Output unit. Could be choosen between :
            B, KB, MB, GB, TB, PB, EB, ZB, KiB, MiB, GiB, TiB, PiB, EiB, ZiB

    Note : Inspired from
    https://web.archive.org/web/20111010015624/http://blogmag.net/blog/read/38/Print_human_readable_file_size # noqa #pylint: disable
    """

    forced_unit = forced_unit.lower()

    if 'i' in forced_unit:
        divisor = 1024.0
        units = ['B', 'KiB', 'MiB', 'GiB', 'TiB', 'PiB', 'EiB', 'ZiB']
    else:
        divisor = 1000.0
        units = ['B', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB']

    lower_units = map(str.lower, units)

    if forced_unit in lower_units or forced_unit == '':
        for unit in units:
            if ((abs(bytes_size) < divisor and forced_unit == '')
                    or forced_unit == unit.lower()):
                if isinstance(bytes_size, int):
                    if bytes_size > 9007199254740992:
                        print(bytes_size)
                        raise ArithmeticError('The bytes_size variable '
                                              '(value : %s) is not convenient '
                                              'with IEEE754 standard. It is '
                                              'not a representable integer, '
                                              'that means it is upper than '
                                              '2^54+1'
                                              % (bytes_size))
                return "%3.1f%s" % (bytes_size, unit)
            bytes_size /= divisor
    return "%.1f%s" % (bytes_size, 'B')


def format_output(output, **kwargs):
    """
    Format the output to have a better display of the collected informations.
        :param output: Dictionnary containing all the collected informations.
        :param **kwargs: any option like size, filter, and so on...
    """
    size = kwargs.get('size', 'B')
    size = size.title()
    output['size'] = str(sizeof_fmt(output['size'], forced_unit=size))

    filter_expression = kwargs.get('filter', None)
    output['filter'] = filter_expression

    return output


def str_to_bucket_list(string_list=None, own_buckets=True):
    """
    Convert a string list into a bucket list.
        :param string_list=[]: A string list
        :param own_buckets=True: If True, include all buckets from S3 resource.

    > str_to_bucket_list([dc-lidar-2015], own_buckets=False)
    >>> [s3.Bucket(name='dc-lidar-2015')]
    > str_to_bucket_list([dc-lidar-2015], own_buckets=True)
    >>> [s3.Bucket(name='testcoveotofull'), s3.Bucket(name='dc-lidar-2015')]
    """

    if string_list is None:
        string_list = []
    bucket_list = []
    if own_buckets:
        bucket_list = list(S3.buckets.all())
    for item in string_list:
        bucket_to_append = S3.Bucket(item)
        if bucket_to_append not in bucket_list:
            bucket_list.append(S3.Bucket(item))

    return bucket_list


def main(**kwargs):
    '''
    Main function
    '''
    force_update = kwargs.get('force_update', False)
    bucket_list = kwargs.get('bucket_list', None)
    default_timeout = kwargs.get('default_timeout', 60)
    filepath = kwargs.get('filepath', default_csv_filepath())
    block_data_crunching = kwargs.get('block_data_crunching', False)
    regex_filtering = kwargs.get('regex_filtering', '^[a-z]*')
    fmt_size = kwargs.get('fmt_size', "B")

    if ((is_file_expired(filepath, default_timeout) or force_update)
            and not block_data_crunching):
        LOGGER.warning("The csv file will be replaced \
        (too old or update forced by user)")
        LOGGER.debug("Csv file location : %s", filepath)
        init_csv_file(filepath)
        LOGGER.info("crunching data from cloud provider")
        crunch_data(filepath, bucket_list)
        LOGGER.info("crunching data ... done")

    infos = collect_informations(filepath, regex_filtering)
    formated_output_list = []
    for info in infos:
        formated_output = format_output(info.copy(),
                                        size=fmt_size,
                                        filter=regex_filtering)
        print(formated_output)
        formated_output_list.append(formated_output.copy())
    return formated_output_list


if __name__ == '__main__':  # pragma: no cover
    import sys
    main(**dict(arg.split('=') for arg in sys.argv[1:]))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.md') as readme_file:
    readme = readme_file.read()

with open('HISTORY.md') as history_file:
    history = history_file.read()

requirements = [
    # TODO: put package requirements here
]

setup_requirements = [
    'pytest-runner',
    # TODO(tofull): put setup requirements (distutils extensions, etc.) here
]

test_requirements = [
    'pytest',
    # TODO: put package test requirements here
]

setup(
    name='s3_analyst',
    version='0.3.1',
    description="An AWS S3 storage analysis tool.",
    long_description=readme + '\n\n' + history,
    author="Loïc Messal",
    author_email='loic.messal@orange.fr',
    url='https://github.com/tofull/devops-coding-challenge',
    packages=find_packages(include=['s3_analyst']),
    include_package_data=True,
    install_requires=requirements,
    license="GNU General Public License v3",
    zip_safe=False,
    keywords='s3_analyst',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
    ],
    test_suite='tests',
    tests_require=test_requirements,
    setup_requires=setup_requirements,
)
